#!/usr/bin/python
# -*- encoding: utf-8 -*-
# Pygmusic: A simple music sequencer in PyGame.  By Kragen Sitaker, 2007-12-07
# 
# Drag things around with the mouse; the right mouse button exits.
# 
# I, the copyright holder of this work, hereby release it into the
# public domain. This applies worldwide.
# 
# In case this is not legally possible, I grant any entity the right
# to use this work for any purpose, without any conditions, unless
# such conditions are required by law.

import pygame, time, os, sys
black = (0, 0, 0)
white = (255, 255, 255)

def padd((x1, y1), (x2, y2)):
    "Add an offset to a 2D point."
    return (x2 + x1, y2 + y1)
def diff((x1, y1), (x2, y2)):
    "Take the offset between two 2D points."
    return (x2 - x1, y2 - y1)

halo_images = {}
def halo_image(size):
    "An image for a halo effect to show that something has happened."
    if halo_images.has_key(size): return halo_images[size]
    halosize = size * 2
    new_img = pygame.surface.Surface((halosize, halosize))
    # I wanted to do a circle, but it turns out that in SDL you can't
    # have both a per-pixel alpha and a per-surface alpha, and I
    # figured that fading the per-pixel alpha in a Python nested loop
    # would be too slow, so for now, it's a white square.
    new_img.fill(white)
    halo_images[size] = new_img
    return new_img

def contains((x, y, width, height), (px, py)):
    "Tells whether a rectangle contains a point."
    return (x <= px < x + width and y <= py < y + height)

class Visible:
    """Things that get drawn on the screen and maybe handle mouse clicks
    or make sounds.

    These objects can be added to the world with world.add and are
    expected to be able to do the following:
    - obj.contains(point): tell whether a mouse position is in the object
    - obj.handle_click(world, event): handle a mouse button going down
    - obj.play(world): make any appropriate sounds
    - obj.center(): return (x, y) center to decide when to be played
    - obj.draw(self, world, screen): display itself; called every frame.

    Additionally, if an object is draggable, it needs to support these:
    - obj.move(delta): move by delta x and delta y
    - obj.handle_drop(world): finish being dragged

    """
    def contains(self, point):
        x, y = self.pos
        return contains((x, y, self.width, self.height), point)
    def center(self):
        x, y = self.pos
        return (x + self.width / 2, y + self.height / 2)
    def handle_click(self, world, ev):
        "Default handle_click method does nothing."
        pass
    def play(self, world):
        "Default play method does nothing."
        pass

class Timer(Visible):
    "A horizontal strip on the screen that plays things in it repeatedly."
    def __init__(self, pos, width, height, cycletime, color, active=True):
        "cycletime is the time to go through the whole cycle."
        self.pos = pos
        self.width = width
        self.height = height
        self.cycletime = cycletime
        self.color = color
        self.active = active
        self.start = time.time()
        self.lastoffset = 0
    def offset(self):
        "The current position of the cursor for this timer."
        phase = ((time.time() - self.start) % self.cycletime) / self.cycletime
        return int(phase * self.width + 0.5)
    def drawvbar(self, screen, offset, color):
        "Draws a timing mark at a given offset."
        screen.fill(color, (padd(self.pos, (offset, 0)), (1, self.height)))
    def drawvbars(self, screen, n, color):
        "Draw timing marks for odd intervals of every nth of a cycle."
        for ii in range(1, n, 2):
            self.drawvbar(screen, ii * self.width / float(n), color)
    def draw(self, world, screen):
        screen.fill((self.color, self.color, self.color, 128),
                    (self.pos, (self.width, self.height)))
        self.drawvbars(screen, 2, (192, 192, 192, 128))
        self.drawvbars(screen, 4, (192, 192, 192, 128))
        self.drawvbars(screen, 8, (144, 144, 144, 128))
        self.drawvbars(screen, 16,(136, 136, 136, 128))
        if self.active: self.draw_cursor(world, screen)
    def draw_cursor(self, world, screen):
        """Draws the white box that represents the currently playing period,
        and also plays the sounds found within."""
        offset = self.offset()
        if offset >= self.lastoffset:
            rects = [pygame.Rect((padd(self.pos, (self.lastoffset, 0)),
                                  (offset - self.lastoffset, self.height)))]
        else:
            rects = [pygame.Rect((padd(self.pos, (self.lastoffset, 0)),
                                  (self.width - self.lastoffset, self.height))),
                     pygame.Rect((self.pos, (offset, self.height)))]
        for rect in rects:
            screen.fill((255,255,255,128), rect)
            for obj in world.objects_in(rect): obj.play(world)
        self.lastoffset = offset
    def handle_click(self, world, ev):
        "Turns the timer on and off when clicked."
        if not self.active: self.lastoffset = self.offset()
        self.active = not self.active

class ImageDisplay(Visible):
    "A visible object that displays by merely blitting an image."
    def __init__(self, pos, image):
        (self.pos, self.image) = (pos, image)
        self.width, self.height = image.get_size()
    def draw(self, world, surface):
        "ImageDisplay.draw just blits the object's image."
        surface.blit(self.image, self.pos)

class Halo(ImageDisplay):
    "An image display object that fades to nothing over 0.6 seconds."
    def __init__(self, pos, image):
        # The image.subsurface thing is so that we can have a separate
        # alpha per halo.
        ImageDisplay.__init__(self, pos, image.subsurface(image.get_rect()))
        self.start = time.time()
    def draw(self, world, surface):
        age = time.time() - self.start
        opacity = int(255 * (0.6 - age))
        if opacity <= 0: world.delete(self)
        else:
            self.image.set_alpha(opacity)
            ImageDisplay.draw(self, world, surface)

class SoundBase(ImageDisplay):
    "Visible objects that play sounds inherit from here."
    def play(self, world):
        self.sound.play()
        halo = halo_image(self.height)
        delta = ((self.image.get_width() - halo.get_width()) / 2,
                 (self.image.get_height() - halo.get_height()) / 2)
        world.add_nonclickable(Halo(padd(self.pos, delta), halo))

class SoundSource(SoundBase):
    "Here is a source you can drag new sounds from."
    def __init__(self, pos, image, sound, instance_image):
        "instance_image is the image for instances from this source."
        ImageDisplay.__init__(self, pos, image)
        self.sound = sound
        self.instance_image = instance_image
    def handle_click(self, world, ev):
        "Start a drag with a new instance of this sound."
        # The (15, 15) offset has two purposes:
        # - it makes it clear that the new thing is a new thing, and
        #   not a change of color of the old thing;
        # - it happens to put that thing by default into one of the
        #   timers, so that if you don't drag it anywhere, it will still
        #   get played.
        new = SoundInstance(padd(self.pos, (15, 15)),
                            self.instance_image, self.sound)
        world.add(new)
        new.start_drag(world, ev)
        self.play(world)

class SoundInstance(SoundBase):
    "A draggable sound that you can put in the tracks."
    def __init__(self, pos, image, sound):
        ImageDisplay.__init__(self, pos, image)
        self.sound = sound
    def move(self, delta):
        self.pos = padd(delta, self.pos)
    def start_drag(self, world, ev):
        world.raise_to_top(self)
        world.grab(self, ev.pos)
    def handle_drop(self, world):
        "If we're dropped on the trash, we suicide."
        # we cheat and look a pixel up and left
        droptarget = world.object_at(padd(self.pos, (-1, -1)))
        # ew, isinstance.  But it's simpler for now.
        if isinstance(droptarget, Trash): world.delete(self)
    def handle_click(self, world, ev):
        self.start_drag(world, ev)
        self.play(world)

class Trash(ImageDisplay):
    """The trashcan deletes SoundInstances.  Empty class, because its
    behavior actually lives in SoundInstance.handle_drop."""

class World:
    """Manages the set of stuff you see on the screen and routes
    events.  I think this is basically the Smalltalk MVC Controller."""
    def __init__(self, screen):
        "screen is an SDL/PyGame surface to draw on."
        self.screen = screen
        self.objects = []               # clickable visible objects
        self.nonclickable_objects = []  # halos and such
        self.grab(None, None)           # initialize drag state
    def redraw(self):
        "This gets called whenever there's idle time, i.e. each frame."
        self.screen.fill(black)
        for obj in self.objects + self.nonclickable_objects:
            obj.draw(self, self.screen)
    def add(self, obj):
        "Call to put a new visible, clickable object on the screen."
        self.objects.append(obj)
    def add_nonclickable(self, obj):
        """Like add, but for nonclickable objects above everything.

        I added this for halos because they were accidentally getting
        drawn underneath other screen objects, and they aren't
        supposed to impede the clickability of the objects they're
        haloing around.
        """
        self.nonclickable_objects.append(obj)
    def object_at(self, pos):
        "Return the topmost object at pos, or None."
        for ii in range(len(self.objects)-1, -1, -1):
            if self.objects[ii].contains(pos):
                return self.objects[ii]
    def handle_click(self, ev):
        "Route a click event to the relevant object."
        obj = self.object_at(ev.pos)
        if obj: obj.handle_click(self, ev)
        else: self.ungrab()
    def handle_release(self, ev):
        "Handle a button release event."
        self.ungrab()
    def handle_motion(self, ev):
        "Handle mouse motion, by doing a mouse drag if needed."
        if self.dragobj is None: return
        self.dragobj.move(diff(self.dragpos, ev.pos))
        self.dragpos = ev.pos
    def ungrab(self):
        "Terminate any drag."
        if self.dragobj is not None: self.dragobj.handle_drop(self)
        self.grab(None, None)
    def grab(self, obj, pos):
        "Start dragging some object."
        (self.dragobj, self.dragpos) = (obj, pos)
    def delete(self, obj):
        "Remove a visible object (clickable or not) from the world."
        if obj in self.nonclickable_objects:
            self.nonclickable_objects.remove(obj)
        else:
            self.objects.remove(obj)
    def raise_to_top(self, obj):
        "Move an object to the top of the drawing stack."
        self.delete(obj)
        self.add(obj)
    def objects_in(self, arect):
        """Find all the objects whose centers are in a rectangle.

        This is used by the timer to figure out what to play.
        """
        return [obj for obj in self.objects
                if contains(arect, obj.center())]
            
def main(argv):
    mydir = os.path.split(argv[0])[0]
    pygame.init()
    screen = pygame.display.set_mode((640, 480), pygame.FULLSCREEN)

    world = World(screen)
    def timer(y, cycletime, color, active=True):
        world.add(Timer(pos=(100, y), width=440, height=30,
                      cycletime=cycletime*0.16667, color=color, active=active))

    font = pygame.font.Font(None, 24)   # use default font, 24 pixels high
    def getsound(soundname):
        path = os.path.join(mydir, soundname + '.wav')
        return pygame.mixer.Sound(path)
    def renderletter(letter, color): return font.render(letter, 1, color)
    def addsource(xpos, letter, soundname):
        world.add(SoundSource(pos=(xpos, 80),
                              image=renderletter(letter, (255, 128, 128)),
                              instance_image=renderletter(letter, white),
                              sound=getsound(soundname)))
    def addsound(ypos, letter, soundname):
        world.add(SoundInstance(pos=(95, ypos),
                                image=renderletter(letter, white),
                                sound=getsound(soundname)))
        

    timer(100, 3, 128)
    addsound(100, 'B', 'reflect_brick')
    timer(130, 4, 120)
    addsound(130, 'A', 'reflect_paddle')
    timer(160, 5, 128, active=False)
    addsound(160, 'Z', 'extraball')
    timer(190, 7, 120, active=False)
    addsound(190, 'Y', 'score')

    addsource(100, 'Y', 'score')
    addsource(120, 'Z', 'extraball')
    addsource(140, 'A', 'reflect_paddle')
    addsource(160, 'B', 'reflect_brick')
    addsource(180, 'C', 'menu_click')
    addsource(200, 'D', 'shrink')

    trashf = os.path.join(mydir, 'trashcan_empty.png')
    trashimg = pygame.image.load(trashf).convert()
    world.add(Trash((100, 300), trashimg))

    world.add(ImageDisplay((200, 300),
        font.render("Drag things around with the mouse.", 1, white)))
    world.add(ImageDisplay((200, 320),
        font.render("The right mouse button exits.", 1, white)))

    frames = 0
    start = time.time()
    while 1:
        ev = pygame.event.poll()
        if ev.type == pygame.NOEVENT:
            world.redraw()
            frames += 1
            pygame.display.flip()
            pygame.time.wait(0) # ms
        elif ev.type == pygame.MOUSEMOTION:
            world.handle_motion(ev)
        elif ev.type == pygame.MOUSEBUTTONDOWN:
            if ev.button == 3: break
            world.handle_click(ev)
        elif ev.type == pygame.MOUSEBUTTONUP:
            world.handle_release(ev)
        elif ev.type == pygame.QUIT: break
    end = time.time()
    print "%.2f seconds, %.2f fps" % ((end - start), frames / (end - start))

if __name__ == '__main__': main(sys.argv)
